@@ -1,16 +1,12 @@
 """
 Container-/OPF-based input OEBBook reader.
 """
-from __future__ import with_statement
-from __future__ import print_function
+
 
 __license__   = 'GPL v3'
 __copyright__ = '2008, Marshall T. Vandegrift <llasram@gmail.com>'
 
-import sys, os, uuid, copy, re, cStringIO
-from itertools import izip
-from urlparse import urldefrag, urlparse
-from urllib import unquote as urlunquote
+import sys, os, uuid, copy, re, io
 from collections import defaultdict
 
 from lxml import etree
@@ -25,11 +21,14 @@ from calibre.ebooks.oeb.base import namespace, barename, XPath, xpath, \
                                     urlnormalize, BINARY_MIME, \
                                     OEBError, OEBBook, DirContainer
 from calibre.ebooks.oeb.writer import OEBWriter
+from calibre.utils.xml_parse import safe_xml_fromstring
 from calibre.utils.cleantext import clean_xml_chars
 from calibre.utils.localization import get_lang
 from calibre.ptempfile import TemporaryDirectory
 from calibre.constants import __appname__, __version__
 from calibre import guess_type, xml_replace_entities
+from polyglot.builtins import unicode_type, zip
+from polyglot.urllib import unquote, urldefrag, urlparse
 
 __all__ = ['OEBReader']
 
@@ -110,23 +109,18 @@ class OEBReader(object):
         data = re.sub(r'http://openebook.org/namespaces/oeb-package/1.0(/*)',
                 OPF1_NS, data)
         try:
-            opf = etree.fromstring(data)
+            opf = safe_xml_fromstring(data)
         except etree.XMLSyntaxError:
             data = xml_replace_entities(clean_xml_chars(data), encoding=None)
             try:
-                opf = etree.fromstring(data)
+                opf = safe_xml_fromstring(data)
                 self.logger.warn('OPF contains invalid HTML named entities')
             except etree.XMLSyntaxError:
                 data = re.sub(r'(?is)<tours>.+</tours>', '', data)
                 data = data.replace('<dc-metadata>',
                     '<dc-metadata xmlns:dc="http://purl.org/metadata/dublin_core">')
-                try:
-                    opf = etree.fromstring(data)
-                    self.logger.warn('OPF contains invalid tours section')
-                except etree.XMLSyntaxError:
-                    from calibre.ebooks.oeb.parse_utils import RECOVER_PARSER
-                    opf = etree.fromstring(data, parser=RECOVER_PARSER)
-                    self.logger.warn('OPF contains invalid markup, trying to parse it anyway')
+                opf = safe_xml_fromstring(data)
+                self.logger.warn('OPF contains invalid tours section')
 
         ns = namespace(opf.tag)
         if ns not in ('', OPF1_NS, OPF2_NS):
@@ -137,7 +131,7 @@ class OEBReader(object):
     def _metadata_from_opf(self, opf):
         from calibre.ebooks.metadata.opf2 import OPF
         from calibre.ebooks.oeb.transforms.metadata import meta_info_to_oeb_metadata
-        stream = cStringIO.StringIO(etree.tostring(opf, xml_declaration=True, encoding='utf-8'))
+        stream = io.BytesIO(etree.tostring(opf, xml_declaration=True, encoding='utf-8'))
         o = OPF(stream)
         pwm = o.primary_writing_mode
         if pwm:
@@ -151,7 +145,7 @@ class OEBReader(object):
                 dict(a=__appname__, v=__version__)
         meta_info_to_oeb_metadata(mi, self.oeb.metadata, self.logger)
         m = self.oeb.metadata
-        m.add('identifier', str(uuid.uuid4()), id='uuid_id', scheme='uuid')
+        m.add('identifier', unicode_type(uuid.uuid4()), id='uuid_id', scheme='uuid')
         self.oeb.uid = self.oeb.metadata.identifier[-1]
         if not m.title:
             m.add('title', self.oeb.translate(__('Unknown')))
@@ -185,7 +179,7 @@ class OEBReader(object):
         return bad
 
     def _manifest_add_missing(self, invalid):
-        import cssutils
+        import css_parser
         manifest = self.oeb.manifest
         known = set(manifest.hrefs)
         unchecked = set(manifest.values())
@@ -199,8 +193,8 @@ class OEBReader(object):
                     try:
                         data = item.data
                     except:
-                        self.oeb.log.exception(u'Failed to read from manifest '
-                                u'entry with id: %s, ignoring'%item.id)
+                        self.oeb.log.exception('Failed to read from manifest '
+                                'entry with id: %s, ignoring'%item.id)
                         invalid.add(item)
                         continue
                 if data is None:
@@ -225,7 +219,7 @@ class OEBReader(object):
                             new.add(href)
                 elif item.media_type in OEB_STYLES:
                     try:
-                        urls = list(cssutils.getUrls(data))
+                        urls = list(css_parser.getUrls(data))
                     except:
                         urls = []
                     for url in urls:
@@ -235,7 +229,7 @@ class OEBReader(object):
                         if not scheme and href not in known:
                             new.add(href)
             unchecked.clear()
-            warned = set([])
+            warned = set()
             for href in new:
                 known.add(href)
                 is_invalid = False
@@ -277,13 +271,13 @@ class OEBReader(object):
                 media_type = media_type.lower()
             fallback = elem.get('fallback')
             if href in manifest.hrefs:
-                self.logger.warn(u'Duplicate manifest entry for %r' % href)
+                self.logger.warn('Duplicate manifest entry for %r' % href)
                 continue
             if not self.oeb.container.exists(href):
-                self.logger.warn(u'Manifest item %r not found' % href)
+                self.logger.warn('Manifest item %r not found' % href)
                 continue
             if id in manifest.ids:
-                self.logger.warn(u'Duplicate manifest id %r' % id)
+                self.logger.warn('Duplicate manifest id %r' % id)
                 id, href = manifest.generate(id, href)
             manifest.add(id, href, media_type, fallback)
         invalid = self._manifest_prune_invalid()
@@ -320,7 +314,7 @@ class OEBReader(object):
             unchecked = new
         version = int(self.oeb.version[0])
         removed_items_to_ignore = getattr(self.oeb, 'removed_items_to_ignore', ())
-        for item in sorted(extras):
+        for item in extras:
             if item.href in removed_items_to_ignore:
                 continue
             if version >= 2:
@@ -334,7 +328,7 @@ class OEBReader(object):
         for elem in xpath(opf, '/o2:package/o2:spine/o2:itemref'):
             idref = elem.get('idref')
             if idref not in manifest.ids:
-                self.logger.warn(u'Spine item %r not found' % idref)
+                self.logger.warn('Spine item %r not found' % idref)
                 continue
             item = manifest.ids[idref]
             if item.media_type.lower() in OEB_DOCS and hasattr(item.data, 'xpath') and not getattr(item.data, 'tag', '').endswith('}ncx'):
@@ -366,7 +360,7 @@ class OEBReader(object):
                         corrected_href = href
                         break
                 if corrected_href is None:
-                    self.logger.warn(u'Guide reference %r not found' % ref_href)
+                    self.logger.warn('Guide reference %r not found' % ref_href)
                     continue
                 ref_href = corrected_href
             typ = elem.get('type')
@@ -429,7 +423,7 @@ class OEBReader(object):
                     'descendant::calibre:meta[@name = "description"]')
             if descriptionElement:
                 description = etree.tostring(descriptionElement[0],
-                method='text', encoding=unicode).strip()
+                method='text', encoding='unicode').strip()
                 if not description:
                     description = None
             else:
@@ -454,7 +448,7 @@ class OEBReader(object):
         ncx = item.data
         title = ''.join(xpath(ncx, 'ncx:docTitle/ncx:text/text()'))
         title = COLLAPSE_RE.sub(' ', title.strip())
-        title = title or unicode(self.oeb.metadata.title[0])
+        title = title or unicode_type(self.oeb.metadata.title[0])
         toc = self.oeb.toc
         toc.title = title
         navmaps = xpath(ncx, 'ncx:navMap')
@@ -541,7 +535,7 @@ class OEBReader(object):
         use = titles
         if len(titles) > len(set(titles)):
             use = headers
-        for title, item in izip(use, self.oeb.spine):
+        for title, item in zip(use, self.oeb.spine):
             if not item.linear:
                 continue
             toc.add(title, item.href)
@@ -631,17 +625,17 @@ class OEBReader(object):
         with TemporaryDirectory('_html_cover') as tdir:
             writer = OEBWriter()
             writer(self.oeb, tdir)
-            path = os.path.join(tdir, urlunquote(hcover.href))
+            path = os.path.join(tdir, unquote(hcover.href))
             data = render_html_svg_workaround(path, self.logger)
             if not data:
-                data = ''
+                data = b''
         id, href = self.oeb.manifest.generate('cover', 'cover.jpg')
         item = self.oeb.manifest.add(id, href, JPEG_MIME, data=data)
         return item
 
     def _locate_cover_image(self):
         if self.oeb.metadata.cover:
-            id = unicode(self.oeb.metadata.cover[0])
+            id = unicode_type(self.oeb.metadata.cover[0])
             item = self.oeb.manifest.ids.get(id, None)
             if item is not None and item.media_type in OEB_IMAGES:
                 return item
