@@ -1,27 +1,29 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python
 # vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
-from __future__ import with_statement
-from __future__ import print_function
+
 
 __license__   = 'GPL v3'
 __copyright__ = '2009, Kovid Goyal <kovid@kovidgoyal.net>'
 __docformat__ = 'restructuredtext en'
 
-import sys, os, re
-from xml.sax.saxutils import escape
+import os
+import re
+import sys
 from string import Formatter
+from xml.sax.saxutils import escape
 
-from lxml import etree
-
-from calibre import guess_type, strftime
+from calibre import guess_type, prepare_string_for_xml, strftime
 from calibre.constants import iswindows
-from calibre.ebooks.oeb.base import XPath, XHTML_NS, XHTML, xml2text, urldefrag, urlnormalize
-from calibre.library.comments import comments_to_html, markdown
-from calibre.utils.date import is_date_undefined, as_local_time
-from calibre.utils.icu import sort_key
 from calibre.ebooks.chardet import strip_encoding_declarations
 from calibre.ebooks.metadata import fmt_sidx, rating_to_stars
-from polyglot.builtins import unicode_type, map
+from calibre.ebooks.metadata.sources.identify import urls_from_identifiers
+from calibre.ebooks.oeb.base import (
+    XHTML, XHTML_NS, XPath, urldefrag, urlnormalize, xml2text
+)
+from calibre.library.comments import comments_to_html, markdown
+from calibre.utils.date import as_local_time, is_date_undefined
+from calibre.utils.icu import sort_key
+from polyglot.builtins import map, unicode_type
 
 JACKET_XPATH = '//h:meta[@name="calibre-content" and @content="jacket"]'
 
@@ -188,7 +190,7 @@ class Series(unicode_type):
             combined = roman = escape(series or u'')
         s = unicode_type.__new__(self, combined)
         s.roman = roman
-        s.name = escape(series or u'')
+        s.name = escape(series or '')
         s.number = escape(fmt_sidx(series_index or 1.0, use_roman=False))
         s.roman_number = escape(fmt_sidx(series_index or 1.0, use_roman=True))
         return s
@@ -232,6 +234,30 @@ def postprocess_jacket(root, output_profile, has_data):
         extract_class('cbj_kindle_banner_hr')
 
 
+class Attributes:
+
+    def __getattr__(self, name):
+        return 'none'
+
+
+class Identifiers:
+
+    def __init__(self, idents):
+        self.identifiers = idents or {}
+        self.display = Attributes()
+        for k in self.identifiers:
+            setattr(self.display, k, 'initial')
+        links = []
+        for x in urls_from_identifiers(self.identifiers):
+            name, id_typ, id_val, url = (prepare_string_for_xml(e, True) for e in x)
+            links.append(f'<a href="{url}" title="{id_typ}:{id_val}">{name}</a>')
+        self.links = ', '.join(links)
+        self.display.links = 'initial' if self.links else 'none'
+
+    def __getattr__(self, name):
+        return self.identifiers.get(name, '')
+
+
 def render_jacket(mi, output_profile,
         alt_title=_('Unknown'), alt_tags=[], alt_comments='',
         alt_publisher='', rescale_fonts=False, alt_authors=None):
@@ -260,7 +286,7 @@ def render_jacket(mi, output_profile,
             pubdate = ''
         else:
             dt = as_local_time(mi.pubdate)
-            pubdate = strftime(u'%Y', dt.timetuple())
+            pubdate = strftime('%Y', dt.timetuple())
     except:
         pubdate = ''
 
@@ -270,7 +296,6 @@ def render_jacket(mi, output_profile,
 
     comments = mi.comments if mi.comments else alt_comments
     comments = comments.strip()
-    orig_comments = comments
     if comments:
         comments = comments_to_html(comments)
 
@@ -286,20 +311,23 @@ def render_jacket(mi, output_profile,
     has_data = {}
 
     def generate_html(comments):
+        display = Attributes()
         args = dict(xmlns=XHTML_NS,
-                    title_str=title_str,
-                    css=css,
-                    title=title,
-                    author=author,
-                    publisher=publisher,
-                    pubdate_label=_('Published'), pubdate=pubdate,
-                    series_label=_('Series'), series=series,
-                    rating_label=_('Rating'), rating=rating,
-                    tags_label=_('Tags'), tags=tags,
-                    comments=comments,
-                    footer='',
-                    searchable_tags=' '.join(escape(t)+'ttt' for t in tags.tags_list),
-                    )
+            title_str=title_str,
+            identifiers=Identifiers(mi.identifiers),
+            css=css,
+            title=title,
+            author=author,
+            publisher=publisher,
+            pubdate_label=_('Published'), pubdate=pubdate,
+            series_label=ngettext('Series', 'Series', 1), series=series,
+            rating_label=_('Rating'), rating=rating,
+            tags_label=_('Tags'), tags=tags,
+            comments=comments,
+            footer='',
+            display=display,
+            searchable_tags=' '.join(escape(t)+'ttt' for t in tags.tags_list),
+        )
         for key in mi.custom_field_keys():
             m = mi.get_user_metadata(key, False) or {}
             try:
@@ -326,6 +354,7 @@ def render_jacket(mi, output_profile,
                 else:
                     args[dkey] = escape(val)
                 args[dkey+'_label'] = escape(display_name)
+                setattr(display, dkey, 'none' if mi.is_null(key) else 'initial')
             except Exception:
                 # if the val (custom column contents) is None, don't add to args
                 pass
@@ -340,27 +369,25 @@ def render_jacket(mi, output_profile,
         # Don't change this unless you also change it in template.xhtml
         args['_genre_label'] = args.get('_genre_label', '{_genre_label}')
         args['_genre'] = args.get('_genre', '{_genre}')
-
-        formatter = SafeFormatter()
-        generated_html = formatter.format(template, **args)
         has_data['series'] = bool(series)
         has_data['tags'] = bool(tags)
         has_data['rating'] = bool(rating)
         has_data['pubdate'] = bool(pubdate)
+        for k, v in has_data.items():
+            setattr(display, k, 'initial' if v else 'none')
+        display.title = 'initial'
+        if mi.identifiers:
+            display.identifiers = 'initial'
+
+        formatter = SafeFormatter()
+        generated_html = formatter.format(template, **args)
 
         return strip_encoding_declarations(generated_html)
 
-    from calibre.ebooks.oeb.base import RECOVER_PARSER
+    from calibre.ebooks.oeb.polish.parsing import parse
+    raw = generate_html(comments)
+    root = parse(raw, line_numbers=False, force_html5_parse=True)
 
-    try:
-        root = etree.fromstring(generate_html(comments), parser=RECOVER_PARSER)
-    except:
-        try:
-            root = etree.fromstring(generate_html(escape(orig_comments)),
-                parser=RECOVER_PARSER)
-        except:
-            root = etree.fromstring(generate_html(''),
-                parser=RECOVER_PARSER)
     if rescale_fonts:
         # We ensure that the conversion pipeline will set the font sizes for
         # text in the jacket to the same size as the font sizes for the rest of
@@ -369,11 +396,11 @@ def render_jacket(mi, output_profile,
         # the same as for text in the main book. So text with size x em will
         # be rescaled to the same value in both the jacket and the main content.
         #
-        # We cannot use calibre_rescale_100 on the body tag as that will just
+        # We cannot use data-calibre-rescale 100 on the body tag as that will just
         # give the body tag a font size of 1em, which is useless.
         for body in root.xpath('//*[local-name()="body"]'):
             fw = body.makeelement(XHTML('div'))
-            fw.set('class', 'calibre_rescale_100')
+            fw.set('data-calibre-rescale', '100')
             for child in body:
                 fw.append(child)
             body.append(fw)
